@@ -17,15 +17,15 @@ from enum import Enum
 from urllib.parse import urlparse
 
 import io
+import re
 import requests
 import time
+from jinja2 import Environment
 from django.core.files.base import File
 from django.utils.translation import gettext as _
 
 from EPCPyYes.core.v1_2 import events
-from EPCPyYes.core.v1_2.events import Action
 from EPCPyYes.core.v1_2 import template_events
-from EPCPyYes.core.v1_2.CBV import BusinessSteps, Disposition
 from quartet_capture import models, rules, errors as capture_errors
 from quartet_capture.rules import RuleContext
 from quartet_capture.tasks import create_and_queue_task
@@ -36,8 +36,10 @@ from quartet_output import errors
 from quartet_output.models import EPCISOutputCriteria, EndPoint
 from quartet_output.parsing import SimpleOutputParser, BusinessOutputParser
 from quartet_output.transport.http import HttpTransportMixin
-from quartet_output.transport.sftp import SftpTransportMixin
+from quartet_output.transport.tcp import SocketTransportMixin
 from quartet_output.transport.mail import MailMixin
+from quartet_output.transport.sftp import SftpTransportMixin
+from quartet_templates.models import Template
 
 
 class ContextKeys(Enum):
@@ -92,6 +94,67 @@ class ContextKeys(Enum):
     CREATED_TASK_NAME_KEY = 'CREATED_TASK_NAME'
 
 
+class DynamicTemplateMixin:
+    """
+    In certain circumstances you may need to override the default templates
+    used by EPCPyYes during output to handle oddities in other systems
+    implementation of the EPCIS protocol.  Use this mixin in any steps
+    to provide either a path to a new template and/or the name of a
+    QU4RTET Template to use.  QU4RTET templates can be configured via
+    the `quartet_template` module's functionality.
+    """
+
+    def get_template(self, env: Environment, default: str):
+        """
+        Allows implementing steps to load a template from a Template parameter.
+        The template parameter would be the name of a given quartet_templates.model.Template
+        instance.
+        :param env: The jinja template Environment object.
+        :param default: The default path to use if there was not a defined
+            quartet_templates.models.Template instance specified in the
+            step's `Template` step parameter.  This would be the django
+            path to a valid template in any python package on the python
+            path.
+        :return: If no template parameter was specified, will return the
+            template from the default path specified.  If neither could
+            be found an exception will be thrown.
+        """
+        template_name = self.get_parameter('Template', None)
+        if template_name:
+            template_model = Template.objects.get(name=template_name)
+            template = env.from_string(template_model.content)
+        else:
+            template = env.get_template(default)
+        return template
+
+
+class FilterEPCsMixin:
+    """
+    Use this mixin to filter or transform EPCs in certain events by supplying a
+    regular expression.  The most common scenario is to remove EPCs from
+    a given event to reduce redundancy in certain business scenarios.
+    """
+
+    def filter(self, epcs: list, search_value: str, reverse=False) -> list:
+        """
+        Uses a search value to filter out EPCs in a list.  To filter out
+        EPCs, provide a list and a regular expression and any epcs that match
+        will be removed from the list.  To only include EPCs that match your
+        regular expression, set the reverse parameter to True.
+        :param epcs: The list of EPCs to filter.
+        :param search_value: The search value to match with.
+        :param reverse: If this is set to True, the function will only return
+        items that match the search value as opposed to removing items
+        that match.
+        :return: Returns a new list of EPCs with the filter logic applied.
+        """
+        if not reverse:
+            result = [epc for epc in epcs if search_value not in epc]
+        else:
+            result = [epc for epc in epcs if search_value in epc]
+        return result
+
+
 class OutputParsingStep(EPCISParsingStep):
     """
     A Step that can invokes the `quartet_output.EPCISOutputCriteria`
@@ -151,18 +214,14 @@ class OutputParsingStep(EPCISParsingStep):
         # get the parser to use from the parameter value.
         # the loose_enforcement parameter is from the base class
         # `EPCISParsingStep` and determines which parser to use.
-        parser_type = self.get_parser_type()
+        self.get_or_create_parameter(
+            'Skip Parsing',
+            'False',
+            'Whether or not to skip the parsing phase and just filter events.')
+        skip_parsing = self.get_boolean_parameter('Skip Parsing', False)
+        parser_type = self.get_parser_type(skip_parsing)
         self.info('Parser Type %s', str(parser_type))
-        try:
-            if isinstance(data, File):
-                parser = parser_type(data,
-                                     self.epc_output_criteria)
-            else:
-                parser = parser_type(io.BytesIO(data),
-                                     self.epc_output_criteria)
-        except TypeError:
-            parser = SimpleOutputParser(io.BytesIO(data.encode()),
-                                        self.epc_output_criteria)
+        parser = self.instantiate_parser(data, parser_type, skip_parsing)
         self.info(_('Parsing the document...'))
         parser.parse()
         self.info(_('Parsing complete.  %s matching events were found.') %
@@ -170,15 +229,54 @@ class OutputParsingStep(EPCISParsingStep):
         rule_context.context[
             ContextKeys.FILTERED_EVENTS_KEY.value] = parser.filtered_events
 
-    def get_parser_type(self):
+    def instantiate_parser(self, data, parser_type, skip_parsing):
+        """
+        Overide to gain access to the parser before or after parsing.
+        :param data: The data to parse
+        :param parser_type: The type of parser to instantiate.
+        :param skip_parsing: Whether or not to parse the data.
+        :return: The instantiated parser type.
+        """
+        try:
+            if isinstance(data, File):
+                parser = parser_type(data,
+                                     self.epc_output_criteria,
+                                     skip_parsing=skip_parsing)
+            elif isinstance(data, str):
+                parser = parser_type(io.BytesIO(str.encode(data)),
+                                     self.epc_output_criteria,
+                                     skip_parsing=skip_parsing)
+            else:
+                parser = parser_type(io.BytesIO(data),
+                                     self.epc_output_criteria,
+                                     skip_parsing=skip_parsing)
+        except TypeError as te:
+            self.info(str(te))
+            parser = SimpleOutputParser(io.BytesIO(data.encode()),
+                                        self.epc_output_criteria)
+        return parser
+
+    def get_parser_type(self, skip_parsing):
         """
         Override to provide a different parser type.
         :return: The `type` of parser to use.
         """
-        parser_type = SimpleOutputParser if self.loose_enforcement \
-            else BusinessOutputParser
+        if not skip_parsing:
+            parser_type = SimpleOutputParser if self.loose_enforcement \
+                else BusinessOutputParser
+        else:
+            self.info('Useing the internal null parser.')
+            parser_type = self.NullParser
         return parser_type
 
+    class NullParser:
+        def __init__(self, *args, **kwargs) -> None:
+            super().__init__()
+            self.filtered_events = []
+
+        def parse(self, *args, **qwargs):
+            pass
+
 
 class FilteredEventStepMixin:
     """
@@ -562,6 +660,20 @@ class CreateOutputTaskStep(rules.Step):
     the *Endpoint* defined in the output criteria.
     '''
 
+    def __init__(self, db_task: models.Task, **kwargs):
+        super().__init__(db_task, **kwargs)
+        try:
+            self.run_immediately = self.get_boolean_parameter(
+                'run-immediately', raise_exception=True
+            )
+        except rules.Step.ParameterNotFoundError:
+            self.run_immediately = self.get_or_create_parameter(
+                'Run Immediately',
+                'False',
+                'Whether or not to execute output '
+                'tasks immediately.'
+            ).lower() in ('true',)
+
     def execute(self, data, rule_context: RuleContext):
         '''
         Checks the context for any data under the OUTBOUND_EPCIS_MESSAGE_KEY
@@ -576,8 +688,7 @@ class CreateOutputTaskStep(rules.Step):
             []
         )
         # see if we are just going to forward the inbound data or not
-        forward_data = self.get_boolean_parameter('Forward Data', False) and \
-                       len(filtered_events) > 0
+        forward_data = self.get_boolean_parameter('Forward Data', False)
         if not forward_data:
             # check the context to see if we have any epcis data to send
             self.info(_('Checking the rule context for any data under the '
@@ -615,8 +726,6 @@ class CreateOutputTaskStep(rules.Step):
             # to the rule engine to create a task to process the message
             # using the rule specified in this step's 'Output Rule' parameter
             # create it in a waiting state until after parameters are supplied
-            run_immediately = self.get_boolean_parameter('run-immediately',
-                                                         default=False)
             if forward_data:
                 self.info(_('The Forward Data parameter was specified. The '
                             'step will send the inbound, unmodifed data to '
@@ -625,7 +734,7 @@ class CreateOutputTaskStep(rules.Step):
                 data, output_rule_name,
                 'Output',
                 task_parameters=[task_param],
-                run_immediately=run_immediately
+                run_immediately=self.run_immediately
             )
             rule_context.context[ContextKeys.CREATED_TASK_NAME_KEY] = task.name
             self.info('Created a new output task %s with rule %s',
@@ -645,7 +754,7 @@ class CreateOutputTaskStep(rules.Step):
 
 
 class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin,
-                    MailMixin):
+                    MailMixin, SocketTransportMixin):
     '''
     Uses the transport information within the `EPCISOutputCriteria` placed
     on the context under the EPCIS_OUTPUT_CRITERIA_KEY to send any data that
@@ -760,6 +869,13 @@ class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin,
                 file_extension=self.get_parameter('file-extension', 'txt'),
                 mimetype=self.get_parameter('content-type', 'text/plain')
             )
+        elif protocol.lower() == 'socket':
+            self.socket_send(
+                data,
+                rule_context,
+                output_criteria,
+                self.info
+            )
 
     def _supports_protocol(self, endpoint: EndPoint):
         '''
@@ -773,7 +889,8 @@ class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin,
         parse_result = urlparse(
             endpoint.urn
         )
-        if parse_result.scheme.lower() in ['http', 'https', 'sftp', 'mailto']:
+        if parse_result.scheme.lower() in ['http', 'https', 'sftp', 'mailto',
+                                                                    'socket']:
             return parse_result.scheme
         else:
             raise errors.ProtocolNotSupportedError(_(
@@ -872,7 +989,8 @@ class EPCPyYesFilteredEventOutputStep(rules.Step, FilteredEventStepMixin):
         pass
 
 
-class AppendCommissioningStep(rules.Step, FilteredEventStepMixin):
+class AppendCommissioningStep(rules.Step, FilteredEventStepMixin,
+                              FilterEPCsMixin):
     """
     Will take all of the unique epcs found in the filtered events
     and create an EPCPyYes commissioning event for those epcs and append it to
