@@ -434,7 +434,7 @@ class TestQuartetOutput(TestCase):
 
     def _create_endpoint(self):
         ep = models.EndPoint()
-        ep.urn = 'http://testhost'
+        ep.urn = getattr(settings, 'TEST_SERVER', 'http://testhost')
         ep.name = 'Test EndPoint'
         ep.save()
         return ep
